# -*- coding: utf-8 -*-

import logging
import re
from typing import TYPE_CHECKING

from .text_utils import (
	fixUtf8,
)

if TYPE_CHECKING:
	from typing import Optional

	from .glossary_type import EntryType, GlossaryType


log = logging.getLogger("pyglossary")


class EntryFilter(object):
	name = ""
	desc = ""
	falseComment = ""

	def __init__(self, glos: "GlossaryType") -> None:
		self.glos = glos

	def prepare(self) -> None:
		"""
			run this after glossary info is set and ready
		"""

	def run(self, entry: "EntryType") -> "Optional[EntryType]":
		"""
			returns an Entry object, or None to skip
				may return the same `entry`,
				or modify and return it,
				or return a new Entry object
		"""
		return entry


class TrimWhitespaces(EntryFilter):
	name = "trim_whitespaces"
	desc = "Remove leading/trailing whitespaces from word(s) and definition"

	def run(self, entry: "EntryType") -> "Optional[EntryType]":
		entry.strip()
		entry.replace("\r", "")
		return entry


class NonEmptyWordFilter(EntryFilter):
	name = "non_empty_word"
	desc = "Skip entries with empty word"

	def run(self, entry: "EntryType") -> "Optional[EntryType]":
		if not entry.s_word:
			return None
		return entry


class NonEmptyDefiFilter(EntryFilter):
	name = "non_empty_defi"
	desc = "Skip entries with empty definition"

	def run(self, entry: "EntryType") -> "Optional[EntryType]":
		if not entry.defi:
			return None
		return entry


class RemoveEmptyAndDuplicateAltWords(EntryFilter):
	name = "remove_empty_dup_alt_words"
	desc = "Remove empty and duplicate alternate words"

	def run(self, entry: "EntryType") -> "Optional[EntryType]":
		entry.removeEmptyAndDuplicateAltWords()
		if not entry.l_word:
			return None
		return entry


class FixUnicode(EntryFilter):
	name = "utf8_check"
	desc = "Fix Unicode in word(s) and definition"
	falseComment = "Do not fix Unicode in word(s) and definition"

	def run(self, entry: "EntryType") -> "Optional[EntryType]":
		entry.editFuncWord(fixUtf8)
		entry.editFuncDefi(fixUtf8)
		return entry


class LowerWord(EntryFilter):
	name = "lower"
	desc = "Lowercase word(s)"
	falseComment = "Do not lowercase words before writing"

	def __init__(self, glos: "GlossaryType") -> None:
		EntryFilter.__init__(self, glos)
		self._re_word_ref = re.compile('href=["\'](bword://[^"\']+)["\']')

	def lowerWordRefs(self, defi: str) -> str:
		return self._re_word_ref.sub(
			lambda m: m.group(0).lower(),
			defi,
		)

	def run(self, entry: "EntryType") -> "Optional[EntryType]":
		entry.editFuncWord(str.lower)
		entry.editFuncDefi(self.lowerWordRefs)
		return entry


class RTLDefi(EntryFilter):
	name = "rtl"
	desc = "Make definition right-to-left"

	def run(self, entry: "EntryType") -> "Optional[EntryType]":
		entry.editFuncDefi(lambda defi: f'<div dir="rtl">{defi}</div>')
		return entry


class RemoveHtmlTagsAll(EntryFilter):
	name = "remove_html_all"
	desc = "Remove all HTML tags (not their contents) from definition"

	def __init__(self, glos: "GlossaryType") -> None:
		self._p_pattern = re.compile(
			'<p( [^<>]*?)?>(.*?)</p>',
			re.DOTALL,
		)
		self._div_pattern = re.compile(
			'<div( [^<>]*?)?>(.*?)</div>',
			re.DOTALL,
		)
		self._br_pattern = re.compile(
			"<br[ /]*>",
			re.IGNORECASE,
		)

	def run(self, entry: "EntryType") -> "Optional[EntryType]":
		from bs4 import BeautifulSoup

		def fixStr(st: str) -> str:
			st = self._p_pattern.sub("\\2\n", st)
			# if there is </p> left without opening, replace with <br>
			st = st.replace("</p>", "\n")

			st = self._div_pattern.sub("\\2\n", st)
			# if there is </div> left without opening, replace with <br>
			st = st.replace("</div>", "\n")

			st = self._br_pattern.sub("\n", st)
			return BeautifulSoup(st, "lxml").text

		entry.editFuncDefi(fixStr)
		return entry


class RemoveHtmlTags(EntryFilter):
	name = "remove_html"
	desc = (
		"Remove given comma-separated HTML tags"
		" (not their contents) from definition"
	)

	def __init__(self, glos: "GlossaryType", tagsStr: str) -> None:
		import re
		tags = tagsStr.split(",")
		self.glos = glos
		self.tags = tags
		tagsRE = "|".join(self.tags)
		self.pattern = re.compile(f"</?({tagsRE})( [^>]*)?>")

	def run(self, entry: "EntryType") -> "Optional[EntryType]":
		def fixStr(st: str) -> str:
			return self.pattern.sub("", st)

		entry.editFuncDefi(fixStr)
		return entry


# FIXME: It's is not safe to lowercases everything between < and >
# including class name, element ids/names, scripts, <a href="bword://...">
# etc. How can we fix that?
class NormalizeHtml(EntryFilter):
	name = "normalize_html"
	desc = "Normalize HTML tags in definition (WIP)"

	_tags = (
		"a", "font", "i", "b", "u", "p", "sup",
		"div", "span",
		"table", "tr", "th", "td",
		"ul", "ol", "li",
		"img",
		"br", "hr",
	)

	def __init__(self, glos: "GlossaryType") -> None:
		log.info("Normalizing HTML tags")
		self._pattern = re.compile(
			"(" + "|".join([
				fr"</?{tag}[^<>]*?>"
				for tag in self._tags
			]) + ")",
			re.S | re.I,
		)

	def _subLower(self, m: "re.Match") -> str:
		return m.group(0).lower()

	def _fixDefi(self, st: str) -> str:
		return self._pattern.sub(self._subLower, st)

	def run(self, entry: "EntryType") -> "Optional[EntryType]":
		entry.editFuncDefi(self._fixDefi)
		return entry


class SkipDataEntry(EntryFilter):
	name = "skip_resources"
	desc = "Skip resources / data files"

	def run(self, entry: "EntryType") -> "Optional[EntryType]":
		if entry.isData():
			return None
		return entry


class LanguageCleanup(EntryFilter):
	name = "lang"
	desc = "Language-specific cleanup/fixes"

	def __init__(self, glos: "GlossaryType") -> None:
		EntryFilter.__init__(self, glos)
		self._run_func = None  # type: Callable[[EntryType], [Optional[EntryType]]]

	def prepare(self) -> None:
		langCodes = {
			lang.code
			for lang in (self.glos.sourceLang, self.glos.targetLang)
			if lang is not None
		}
		if "fa" in langCodes:
			self._run_func = self.run_fa
			log.info("Using Persian filter")

	def run_fa(self, entry: "EntryType") -> "Optional[EntryType]":
		from pyglossary.persian_utils import faEditStr
		entry.editFuncWord(faEditStr)
		entry.editFuncDefi(faEditStr)
		# RLM = "\xe2\x80\x8f"
		# defi = "\n".join([RLM+line for line in defi.split("\n")])
		# for GoldenDict ^^ FIXME
		return entry

	def run(self, entry: "EntryType") -> "Optional[EntryType]":
		if self._run_func:
			return self._run_func(entry)
		return entry


class TextListSymbolCleanup(EntryFilter):
	"""
	Symbols like ♦ (diamond) ● (black circle) or * (star) are used in some
	plaintext or even html glossaries to represent items of a list
	(like <li> in proper html).
	This EntryFilter cleans up spaces/newlines issues around them.
	"""

	name = "text_list_symbol_cleanup"
	desc = "Text List Symbol Cleanup"

	winNewlinePattern = re.compile("[\r\n]+")
	spacesNewlinePattern = re.compile(" *\n *")
	blocksNewlinePattern = re.compile("♦\n+♦")

	def cleanDefi(self, st: str) -> str:
		st = st.replace("♦  ", "♦ ")
		st = self.winNewlinePattern.sub("\n", st)
		st = self.spacesNewlinePattern.sub("\n", st)

		st = self.blocksNewlinePattern.sub("♦", st)
		if st.endswith("<p"):
			st = st[:-2]
		st = st.strip()
		if st.endswith(","):
			st = st[:-1]

		return st  # noqa: RET504

	def run(self, entry: "EntryType") -> "Optional[EntryType]":
		entry.editFuncDefi(self.cleanDefi)
		return entry


class PreventDuplicateWords(EntryFilter):
	name = "prevent_duplicate_words"
	desc = "Prevent duplicate words"

	def __init__(self, glos: "GlossaryType") -> None:
		EntryFilter.__init__(self, glos)
		self._wordSet = set()

	def run(self, entry: "EntryType") -> "Optional[EntryType]":
		if entry.isData():
			return entry

		wordSet = self._wordSet
		word = entry.s_word

		if word not in wordSet:
			wordSet.add(word)
			return entry

		n = 2
		while f"{word} ({n})" in wordSet:
			n += 1
		word = f"{word} ({n})"

		wordSet.add(word)
		entry._word = word
		# use entry.editFuncWord?

		return entry


class SkipEntriesWithDuplicateHeadword(EntryFilter):
	name = "skip_duplicate_headword"
	desc = "Skip entries with a duplicate headword"

	def __init__(self, glos: "GlossaryType") -> None:
		EntryFilter.__init__(self, glos)
		self._wset = set()

	def run(self, entry: "EntryType") -> "Optional[EntryType]":
		word = entry.l_word[0]
		if word in self._wset:
			return None
		self._wset.add(word)
		return entry


class TrimArabicDiacritics(EntryFilter):
	name = "trim_arabic_diacritics"
	desc = "Trim Arabic diacritics from headword"

	def __init__(self, glos: "GlossaryType") -> None:
		EntryFilter.__init__(self, glos)
		self._pat = re.compile("[\u064b-\u065f]")

	def run(self, entry: "EntryType") -> "Optional[EntryType]":
		words = list(entry.l_word)
		hw = words[0]
		hw_t = self._pat.sub("", hw)
		hw_t = hw_t.replace("\u0622", "\u0627").replace("\u0623", "\u0627")
		if hw_t == hw or not hw_t:
			return entry
		entry._word = [hw_t] + words
		return entry


class ShowProgressBar(EntryFilter):
	name = "progressbar"
	desc = "Progress Bar"

	def __init__(self, glos: "GlossaryType") -> None:
		EntryFilter.__init__(self, glos)
		self._wordCount = -1
		self._wordCountThreshold = 0
		self._lastPos = 0
		self._index = 0

	def run(self, entry: "EntryType") -> "Optional[EntryType]":
		index = self._index
		self._index = index + 1

		if entry is not None and (bp := entry.byteProgress()):
			if bp[0] > self._lastPos + 100_000:
				self.glos.progress(bp[0], bp[1], unit="bytes")
				self._lastPos = bp[0]
			return entry

		if self._wordCount == -1:
			self._wordCount = len(self.glos)
			self._wordCountThreshold = max(1, min(
				500,
				self._wordCount // 200,
			))

		if self._wordCount > 1 and index % self._wordCountThreshold == 0:
			self.glos.progress(index, self._wordCount)

		return entry


class ShowMaxMemoryUsage(EntryFilter):
	name = "max_memory_usage"
	desc = "Show Max Memory Usage"

	def __init__(self, glos: "GlossaryType") -> None:
		EntryFilter.__init__(self, glos)
		self._max_mem_usage = 0

	def run(self, entry: "EntryType") -> "Optional[EntryType]":
		import os

		import psutil
		usage = psutil.Process(os.getpid()).memory_info().rss // 1024
		if usage > self._max_mem_usage:
			self._max_mem_usage = usage
			word = entry.s_word
			if len(word) > 30:
				word = word[:37] + "..."
			log.trace(f"MaxMemUsage: {usage:,}, {word=}")
		return entry

entryFiltersRules = [
	(None, TrimWhitespaces),
	(None, NonEmptyWordFilter),
	(("skip_resources", False), SkipDataEntry),
	(("utf8_check", False), FixUnicode),
	(("lower", False), LowerWord),
	(("skip_duplicate_headword", False), SkipEntriesWithDuplicateHeadword),
	(("trim_arabic_diacritics", False), TrimArabicDiacritics),
	(("rtl", False), RTLDefi),
	(("remove_html_all", False), RemoveHtmlTagsAll),
	(("remove_html", ""), RemoveHtmlTags),
	(("normalize_html", False), NormalizeHtml),
	(None, LanguageCleanup),

	# TODO
	# (("text_list_symbol_cleanup", False), TextListSymbolCleanup),

	(None, NonEmptyWordFilter),
	(None, NonEmptyDefiFilter),
	(None, RemoveEmptyAndDuplicateAltWords),
]
"""
some entry filters are added conditionally (other than with config):
	- ShowProgressBar
	- ShowMaxMemoryUsage
some entry filters are added by calling a Glossary method
	- PreventDuplicateWords
	- RemoveHtmlTagsAll
"""
